import 'dart:convert';

import 'package:flutter/material.dart';
import 'package:flutter_inappwebview/flutter_inappwebview.dart';

import 'connect_webview.dart';
import 'oauth_webview.dart';
import 'utils.dart';

// TODO: Add types and replace Object
typedef OnSuccessCallback = void Function(dynamic data);
typedef OnErrorCallback = void Function(dynamic error);
typedef OnEventCallback = void Function(dynamic payload);

class PluggyConnect extends StatefulWidget {
  PluggyConnect(
      {Key? key,
      this.url = 'https://connect.pluggy.ai',
      required this.connectToken,
      this.includeSandbox = false,
      this.updateItem,
      this.connectorTypes = const [],
      this.connectorIds = const [],
      this.countries = const [],
      this.language = 'pt',
      this.onSuccess,
      this.onError,
      this.onOpen,
      this.onClose,
      this.onEvent})
      : super(key: key) {
    WidgetsFlutterBinding.ensureInitialized();
    _connectUrl = buildConnectUrl(connectToken, updateItem, connectorIds,
        connectorTypes, countries, includeSandbox, language, url);
  }

  final String url;
  final String connectToken;
  final bool includeSandbox;
  final String? updateItem;

  // TODO: Add types
  final List<String> connectorTypes;
  final List<int> connectorIds;

  // TODO: Add types
  final List<String> countries;
  final String language;
  final OnSuccessCallback? onSuccess;
  final OnErrorCallback? onError;
  final VoidCallback? onOpen;
  final VoidCallback? onClose;
  final OnEventCallback? onEvent;
  late final String _connectUrl;

  @override
  State<StatefulWidget> createState() => _PluggyConnectState();
}

class _PluggyConnectState extends State<PluggyConnect> {
  String? _lastTimestamp;
  String? _oauthUrl;
  bool _isWaitingPluggyResponse = false;
  InAppWebViewController? _connectWebViewController;

  _handleConnectUrlUpdate(String query) {
    final queryParams = Uri.splitQueryString(query.replaceFirst('?', ''));
    final events = queryParams['events'];
    final itemId = queryParams['item_id'];
    final timestamp = queryParams['timestamp'];
    final itemStatus = queryParams['item_status'];
    final executionStatus = queryParams['execution_status'];
    final connectorId = queryParams['connector_id'];
    final connectorName = queryParams['connector_name'];
    final error = queryParams['error'];
    final eventType = events?.split(',').last;

    // skip if there is no new events
    if (eventType == null || _lastTimestamp == timestamp) {
      return;
    }

    setState(() {
      _lastTimestamp = timestamp;
    });

    // event item data
    final itemData = {
      'id': itemId,
      'status': itemStatus,
      'executionStatus': executionStatus,
    };

    // event connector data
    final connectorData = {
      'id': connectorId,
      'name': connectorName,
    };

    if (isConnectEventType(eventType)) {
      final Map<String, dynamic> data = {};
      if (isInstitutionSelectedEventType(eventType)) {
        if (connectorId != null && connectorName != null) {
          data['connector'] = connectorData;
        } else {
          data['connector'] = null;
        }
      }
      if (eventsWithItemInPayload(eventType)) {
        data['item'] = itemData;
      }

      widget.onEvent?.call({
        'event': eventType,
        'timestamp': DateTime.now().millisecondsSinceEpoch,
        ...data
      });
      return;
    }

    switch (eventType) {
      case 'OPEN':
        widget.onOpen?.call();
        break;
      case 'CLOSE':
        widget.onClose?.call();
        break;
      case 'ERROR':
        widget.onError?.call({
          'message': error ?? '',
          ...(itemId != null
              ? {
                  'data': {
                    'item': itemData,
                  }
                }
              : {})
        });
        break;
      case 'SUCCESS':
        widget.onSuccess?.call({'item': itemData});
        break;
    }
  }

  _handleOnOauthOpen(String oauthUrl) {
    setState(() {
      _oauthUrl = oauthUrl;
    });
  }

  _handleConnectWebViewCreated(
      InAppWebViewController connectWebViewController) {
    // store connect webView reference, to send messages to it later
    setState(() {
      _connectWebViewController = connectWebViewController;
    });
  }

  _handleOauthCallback(Map<String, dynamic> oauthResult) async {
    // send Oauth callback result as a message to Connect WebView instance
    if (_connectWebViewController == null) {
      // connect webView has not been created yet, can't send message
      return;
    }

    final oauthResultMessage = {
      ...oauthResult,
      'source': 'connector-oauth-login'
    };

    // send payload to Connect
    await sendMessageToWebView(_connectWebViewController, oauthResultMessage);

    // clear oauth webView state
    setState(() {
      _oauthUrl = null;
      _isWaitingPluggyResponse = false;
    });
  }

  _onWaitingPluggyResponseChange(bool value) {
    setState(() {
      _isWaitingPluggyResponse = value;
    });
  }

  @override
  Widget build(BuildContext context) {
    return Stack(
      children: [
        ConnectWebView(
          url: widget._connectUrl,
          onLocationUpdate: _handleConnectUrlUpdate,
          onOauthOpen: _handleOnOauthOpen,
          onWebViewCreated: _handleConnectWebViewCreated,
        ),
        OauthWebView(
          url: _oauthUrl,
          isWaitingPluggyResponse: _isWaitingPluggyResponse,
          onOauthCallback: _handleOauthCallback,
          onWaitingPluggyResponseChange: _onWaitingPluggyResponseChange,
        )
      ],
    );
  }
}
